 /*
 * converts picture data from Bullfrog's VG Buildings Sprites files
 * to a BMP images
 *
 * Compiled normally, this file is a module without main().
 * Compiled with MAIN_XVGSPRB defined, it's a standalone program.
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "bulcommn.h"
#include "xvgsprb.h"

#ifdef MAIN_XVGSPRB

#define BITMAP_HDR 118

int main (int argc, char **argv)
{
    printf("\nBullfrog Engine 2D Buildings Sprites Extractor for VG Map graphic");
    printf("\n(probably works only with Syndicate for Megadrive)");
    printf("\n-------------------------------\n");
    if (argc<3)
    {
        printf("Not enought parameters.\n");
        printf("Usage:\n");
        printf("  %s <indexfile> <rawfile> <palfile>\n","xvgsprb");
        printf("Output: pic????.bmp\n");
    	return 1;
    }

    VGSPRITETILE stile;
    //Opening and reading RAW file
    {
        FILE *rawfp;
        char fname[255];
        sprintf (fname, "%s", argv[2]);
        rawfp = fopen (fname, "rb");
        if (!rawfp)
        {
            printf("Error: Can't open RAW image file %s.\n", fname);
        	return 1;
        }
        stile.datasize=file_length_opened(rawfp)-BITMAP_HDR;
        stile.data=malloc(stile.datasize+4);
        fseek(rawfp, BITMAP_HDR, SEEK_SET);
        fread (stile.data, stile.datasize, 1, rawfp);
        fclose(rawfp);
    }
    
    printf("Loading palette ...");
    //Reading palette file
    char *palette=malloc(256*3);
    {
        char fname[255];
        sprintf (fname, "%s", argv[3]);
        memset(palette+16*3,0,240*3);
        int retcode=read_palette_rgb(palette,fname,16);
        if (retcode!=0)
          switch (retcode)
          {
          case 1:
             printf("\nError: Cannot open PAL file: %s\n",fname);
             return 11;
          case 2:
             printf("\nError: Cannot read %d colors from PAL file: %s\n",16,fname);
             return 12;
          case 3:
             printf("\nError: PAL file is compressed with RNC, please decompress it\n");
             return 13;
          default:
             printf("\nError: Loading palette file %s returned fail code %d\n",fname,retcode);
             return 19;
          }
    }
    printf("Done.\n");

    //Filling structure with information about Tile file content
    stile.subwidth   = 8;
    stile.subheight  = 8;
    stile.subcount_w = 4;
    stile.subcount_h = 2;
    stile.num_tiles  = 256;
    long num_subtiles=stile.num_tiles*stile.subcount_w*stile.subcount_h;
    int tile_size=stile.subwidth*stile.subcount_w*stile.subheight*stile.subcount_h;
    //checking if the Tile file size is right
    //It is hard to say how large the file should be, because every tile
    //is made of subtiles, and each subtile may be used in many tiles.
    unsigned long proper_stile_size=num_subtiles*(stile.subwidth*stile.subheight/2);
    if (stile.datasize<(proper_stile_size/3))
    {
        printf("Warning: RAW file is surprisingly small!\n");
    }
    //Checking for RNC compression
    if (rnc_compressed_buf(stile.data))
    {
        printf("Warning: RAW file seems to be RNC-compressed!\n");
        printf("         Decompress it first with DERNC!\n");
    }
    //Opening and reading index file
    {
        FILE *sidxfp;
        char fname[255];
        sprintf (fname, "%s", argv[1]);
        sidxfp = fopen (fname, "rb");
        if (!sidxfp)
        {
            printf("Error: Can't open sprite index file %s.\n", fname);
        	return 1;
        }
        stile.look_up_table=malloc(sizeof(long)*num_subtiles);
        // Reading subtitle indices making up tile
        {
            int i;
            for( i=0; i < num_subtiles; i++ )
                stile.look_up_table[i]=read_short_be_file(sidxfp);
        }
        fclose(sidxfp);
    }

    int processed=4*num_subtiles;
    //Interpreting files - Looping through tiles
    unsigned char *tile_pix = malloc(tile_size);
    unsigned char *tile_alpha = malloc(tile_size);
    int picnum;
    for (picnum=0; picnum<stile.num_tiles; picnum++)
    {
    	printf ("\rExtracting: tile number %*d", 4, picnum);
        processed+=get_vgsprite_tile(tile_pix, tile_alpha, picnum, stile);
        char fname[20];
    	sprintf (fname, "pic%0*d.bmp", 4, picnum);
       	write_bmp_idx (fname, stile.subwidth*stile.subcount_w, stile.subheight*stile.subcount_h,
            palette, tile_pix, 0, 1, 2, 4);
    }
    free(tile_pix);
    free(tile_alpha);
    //Summary informations
    printf("\nExtraction complete.\n");
    printf("Processed %d of %d bytes of RAW file.\n",processed,stile.datasize);
    int unused=stile.datasize-processed;
    if (unused>=0)
        printf("Bytes skipped: %d\n",unused);
      else  
        printf("Bytes overlapping: %d\n",-unused);
    free(stile.look_up_table);
    free(stile.data);
    return 0;
}

#endif

void free_map_tiles_rawlist(MAPTILELIST *list)
{
    int picnum;
    for (picnum=0; picnum < list->num_tiles; picnum++)
    {
    free(list->tiles[picnum].data);
    free(list->tiles[picnum].alpha);
    }
    list->num_tiles=0;
    free(list->palette);
//    free(list->tiles);
    list->tiles=NULL;
}

/*
int create_map_tiles_rawlist(MAPTILELIST *list,char *sprite_file_name,char *pal_file_name)
// Reads all the titles to MAPTILELIST structure from given files.
// On success, returns 0.
{
    SPRITETILE stile;
    //Opening and reading DAT file
    {
        FILE *datfp;
        datfp = fopen (sprite_file_name, "rb");
        if (!datfp)
        {
        	return VGSPRBLK_CANNOT_OPEN_DAT;
        }
        stile.datasize=file_length_opened(datfp);
        stile.data=malloc(stile.datasize+4);
        fseek(datfp, 0, SEEK_SET);
        fread (stile.data, stile.datasize, 1, datfp);
        fclose(datfp);
    }
    //Reading palette file
    {
        list->palette=malloc(256*3);
        int retcode=read_palette_rgb(list->palette,pal_file_name,256);
        if (retcode!=0)
          switch (retcode)
          {
          case 1:
             return VGSPRBLK_CANNOT_OPEN_PAL;
          default:
             return VGSPRBLK_CONTENT_BAD_PAL;
          }
    }    
    //Filling internal structure with information about Tile file content
    stile.subwidth   = 32;
    stile.subheight  = 16;
    stile.subcount_w = 2;
    stile.subcount_h = 3;
    stile.num_tiles  = 256;
    long num_subtiles=stile.num_tiles*stile.subcount_w*stile.subcount_h;
    int tile_size=stile.subwidth*stile.subcount_w*stile.subheight*stile.subcount_h;
    //checking if the Tile file size is right
    //It is hard to say how large the file should be, because every tile
    //is made of subtiles, and each subtile may be used in many tiles.
    unsigned long proper_stile_size=num_subtiles*4 + num_subtiles*
          (stile.subwidth*stile.subheight * 5/8);
    if (stile.datasize<(proper_stile_size/3))
             return VGSPRBLK_CONTENT_BAD_DAT;
    stile.look_up_table=malloc(sizeof(long)*num_subtiles);
    // Reading Tile file header info
    // (an array of the offsets of all the subtiles)
    {
        int i;
        for( i=0; i < num_subtiles; i++ )
            stile.look_up_table[i]=read_long_be_buf(stile.data+4*i);
    }
    int processed=4*num_subtiles;
    //Setting output structure informations
    list->num_tiles=stile.num_tiles;
    list->tiles=malloc(list->num_tiles*sizeof(MAPTILE *));
    //Interpreting DAT file - Looping through tiles
    int picnum;
    for (picnum=0; picnum<list->num_tiles; picnum++)
    {
        unsigned char *tile_pix = malloc(tile_size);
        unsigned char *tile_alpha = malloc(tile_size);
        list->tiles[picnum].width=stile.subwidth*stile.subcount_w;
        list->tiles[picnum].height=stile.subheight*stile.subcount_h;
        list->tiles[picnum].alpha=tile_alpha;
        list->tiles[picnum].data=tile_pix;
        processed+=get_sprite_tile(tile_pix, tile_alpha, picnum, stile);
    }
    //Freeing temporary structures
    free(stile.look_up_table);
    free(stile.data);
    return 0;
}
*/

int get_vgsprite_tile(char *pix_map,char *alpha_map, int num, VGSPRITETILE src)
// Extracts the whole tile from building sprite file loadied into
// SPRITETILE structure. Fills pix_map and alpha_map tables.
// Requires tile number (usually 0..255).
{
//!!!TODO
    //Number of source buffer bytes processed
    int processed=0;
    //Subtiles count
    long num_subtiles=src.num_tiles*src.subcount_w*src.subcount_h;
    long subtile_dst_size=src.subwidth*src.subheight;
    long subtile_src_size=(src.subwidth*src.subheight/2)
         + (src.subwidth%2)*src.subheight;
    //Title consists of 6 subtitles indexed 0..5.
    unsigned char *subtile_pix = malloc(subtile_dst_size);
    unsigned char *subtile_alpha = malloc(subtile_dst_size);
    //First three are left half of image, next three are right half
    //So they come as:
    //    |-----------------|
    //    |   0    |   3    |
    //    |-----------------|
    //    |   1    |   4    |
    //    |-----------------|
    //    |   2    |   5    |
    //    |-----------------|
    // Now going through subtitles
    int subnum;
    int debug_maxidx=0;
    for (subnum=0;subnum<src.subcount_w*src.subcount_h;subnum++)
    {
        long look_up_item=src.look_up_table[(num*src.subcount_w*src.subcount_h)+subnum];
//        long look_up_item=2*src.look_up_table[(num*src.subcount_w*src.subcount_h)+subnum/2-1];
//        if ((subnum%2) > 0) look_up_item+=1;
        if (look_up_item>debug_maxidx) debug_maxidx=look_up_item;
        int offset=look_up_item*subtile_src_size;
//!!!!!!!!
//        int offset=(num+subnum)*(subtile_src_size);
        if ((offset > 1798*subtile_src_size)||(offset<0))
        {
            printf("(item %d too large)",offset/subtile_src_size);
            fill_sprite_subtile(subtile_pix,subtile_alpha,src.subwidth,src.subheight,0);
        } else
        {
            processed+=get_vgsprite_subtile(subtile_pix,subtile_alpha,
                                          src.data+offset,src.datasize-offset,
                                          src.subwidth,src.subheight);
        }
        int h,w;
        //The starting coordinates of this subtitle inside the title
//        int sub_w0=src.subwidth *(subnum/src.subcount_h);
//        int sub_h0=src.subheight*(subnum%src.subcount_h);
//!!!!!!
        int sub_w0=src.subwidth *(subnum%src.subcount_w);
        int sub_h0=src.subheight*(subnum/src.subcount_w);
        //Now filling title
        for( h=0; h < src.subheight; h++ )
        for( w=0; w < src.subwidth; w++ )
        {
            int map_idx=w+sub_w0 + ((h+sub_h0)*src.subwidth*src.subcount_w);
            pix_map  [map_idx] = subtile_pix  [w + (h*src.subwidth)];
            alpha_map[map_idx] = subtile_alpha[w + (h*src.subwidth)];
        }
    }
    printf("(max index%5d)",debug_maxidx);
    free(subtile_pix);
    free(subtile_alpha);
    return processed;
}

void fill_sprite_subtile(unsigned char *pix_map,unsigned char *alpha_map, int width, int height, int hstart)
{
    int widx;
    int hidx;
    for( hidx=hstart; hidx < height; hidx++ )
    {
        int hpos=hidx*width;
        for( widx=0; widx < width; widx++ )
            {
                pix_map[hpos+widx] = 255;
                alpha_map[hpos+widx]=255;
            }
    }
}

int get_vgsprite_subtile(unsigned char *pix_map,unsigned char *alpha_map, unsigned char *src_buf, int src_bufsize, int width, int height)
// Give it the buffer with subtile from the RAW file, and it will 
// fill the arrays with the image data. Returns number of source buffer bytes processed
{
    int processed=0;
    //Determining the chunk size
    int chunk_size=width/2;
    if ((width%2)!=0) chunk_size++;
    // for every horizontal strip of pict...
    int hidx;
    for( hidx=0; hidx < height; hidx++ )
        {
            //If the buffer is truncated, fill the rest of output
            if (src_bufsize < processed+chunk_size)
            {
                fill_sprite_subtile(pix_map,alpha_map,width,height,hidx);
                return processed;
            }
            int hpos=hidx*width;
            processed+=chunk_size;
            char *dat_chunk=src_buf+chunk_size*hidx;
            int widx;
            for( widx=0; widx < width; widx++ )
            {
                 unsigned char pixel=dat_chunk[widx/2];
                 if (widx%2) pixel=pixel>>4;
                 pixel&=0x0f;
                // check for alpha and fill the buffer
                if( pixel<1 )
                {
                    alpha_map[hpos+widx]=255;
                    pix_map[hpos+widx]=255;
                }
                 else
                {
                    alpha_map[hpos+widx]=0;
                    pix_map[hpos+widx] = pixel;
                }
            }
        }
    return processed;
}
